"""Various useful routines for the init code."""
from __future__ import division, with_statement
import os, os.path, time, sys
from subprocess import check_call, call
from shutdown import poweroff

########################################################################
# Parsing command-line and ofw bootpath options.
def mount_proc_sys():
    check_call(['/bin/mount','/proc']) # mount /proc
    call(['/bin/mount','-a']) # mount /sys, etc.

########################################################################
#### device names and mount locations
SD_DEV = '/dev/mmcblk0p1'
SD_MNT = '/mnt/sd'
USB_DEV = '/dev/sda1'
USB_MNT = '/mnt/usb'

def blk_mounted(device, mnt, fstype='msdos'):
    """Mount a block device."""
    class blk_mgr(object):
        def __enter__(self):
            check_call(['/bin/mount','-t',fstype,'-o','ro',device,mnt])
        def __exit__(self, type, value, traceback):
            call(['/bin/umount',mnt])
    return blk_mgr()

#####################################################
# "Safe" shell invocation.
def safe_sh(sh_str):
    """Invoke 'str' in a subshell, like os.system(str) does, but raise a
    RuntimeError if the exit status is not 0."""
    if os.system(sh_str) != 0:
        raise RuntimeError("Command exited with non-zero exit status.")

#####################################################
# ensure necessary modules are loaded for sd/usb/eth0
_sd_first = True
def sd_init():
    """Ensure necessary modules are loaded for sd."""
    global _sd_first
    # ignore modprobe failures, since older kernels don't have
    # modular sd (trac #7369).
    call(['/sbin/modprobe','sdhci'])
    call(['/sbin/modprobe','mmc_block'])
    if _sd_first:
        _sd_first = False
        time.sleep(3) # CAFE takes a bit to wake up

_usb_first = True
def usb_init():
    """Ensure necessary modules are loaded for usb."""
    global _usb_first
    # ignore modprobe failures, since older kernels don't have
    # modular usb (trac #7113).
    call(['/sbin/modprobe','ohci-hcd'])
    call(['/sbin/modprobe','usb-storage'])
    if _usb_first:
        _usb_first = False
        time.sleep(5) # usb disks take a while to spin up

def qemu_init():
    """Ensure necessary modules are loaded to mount root when running
    under qemu."""
    check_call(['/sbin/modprobe','piix'])
    check_call(['/sbin/modprobe','ide-disk'])

def net_init():
    """Ensure necessary modules are loaded for network access."""
    check_call(['/sbin/modprobe', 'usb8xxx'])
    call(['/sbin/modprobe','ipv6']) # ipv6 is built statically in recent kernels

#####################################################
# mounting and writing to root.

def ofw_partition(bootpath):
    # pull out the bit between the first : and the next ,
    try:
        part = bootpath.split(':',1)[1].split(',',1)[0]
    except IndexError: # shouldn't happen.
        return None
    if len(part) == 0: return None
    if part.startswith('\\'):
        # OFW algorithm uses the "partition information that has already
        # been established" in this case, which means:
        #  - default is to find the first bootable partition if the disk is
        #    partitioned
        #  - if no partitions are marked bootable, use first non-empty
        #    partition, where non-empty means that the partition has some
        #    sort of filesystem type.
        # But that's ugly.  As wmb says, "just tell people that if
        # they want to boot from a secondary partition, they bloody
        # well have to write a boot path that mentions it"!
        return None # like the man says.
    if part[0] in ['0','1','2','3','4','5','6','7','8','9']:
        try:
            return int(part)
        except ValueError:
            return None # shouldn't happen!
    return str(part) # partition label.

global first
first=True
def root_mounted(xo, bootpath):
    """Mount root.

    Intended to be used as 'with root_mounted(): ...'"""
    class Mgr(object):
        def __enter__(self):
            access = 'ro'
            dev = None
            extra = []
            if xo:
                # if we're booting from sd or usb, mount root from there
                p = ofw_partition(bootpath)
                if p == 'boot':  # boot/root/home/security partition layout.
                    p = 'root'   # we're trying to mount 'root'
                if bootpath.startswith('/pci/usb'):
                    usb_init()
                    dev = '/dev/sda'
                    if p is None: p = 1 # use 1st partition by default.
                    if type(p) is int:
                        dev += str(p)
                    else:
                        extra = ['-L', p]
                elif bootpath.startswith('/pci/sd'):
                    sd_init()
                    if type(p) is not int: p = 1 # 1st part default.
                    dev = '/dev/mmcblk0p%d' % p
                else:
                    # when unpartitioned, bootpath like:
                    # /pci/nandflash@c:\boot\vmlinuz//jffs2-file-system:\boot\vmlinuz
                    # when partitioned, expect bootpath like:
                    # /pci/nandflash@c:root,\boot\vmlinuz//jffs2-file-system:\boot\vmlinuz
                    if p is None: p = 0 # unpartitioned by default
                    if type(p) is int:
                        dev = 'mtd%d' % p
                    else:
                        dev = 'mtd:%s' % p
                    extra = ['-t','jffs2']
            else: # we're running under emulation
                # these modules only needed if we're running in qemu
                from stat import S_IFBLK
                qemu_init()
                if not os.path.exists('/dev/hda1'):
                    os.mknod("/dev/hda1", S_IFBLK | 0777, os.makedev(3, 1))
                dev = '/dev/hda1'
            check_call(['/bin/mount','-n','-o',access]+extra+[dev,'/sysroot'])
        def __exit__(self, type, value, traceback):
            global first
            if first:
                time.sleep(1)
                print "Unmounting root (this may take a while)...",
                sys.stdout.flush()
            check_call(['/bin/umount','/sysroot'])
            if first:
                print "done."
                first = False
    return Mgr()

def writable_mnt(path):
    """Remount a filesystem writable (if it is not already).

    Intended to be used as 'with writable_mnt('/path'): ...'"""
    class Mgr(object):
        def __enter__(self):
            check_call(['/bin/mount','-o','remount,rw',path])
        def __exit__(self, type, value, traceback):
            check_call(['/bin/mount','-o','remount,ro',path])
    return Mgr()

def writable_root():
    """Remount the root filesystem writable (if it is not already).

    Intended to be used as 'with writable_root(): ...'"""
    return writable_mnt('/sysroot')

###############################################################
# OFW device tree I/O

def exists_ofw(dev_path):
    """Returns true if the OFW device tree node exists.  Else returns false."""
    return os.path.exists(os.path.join('/ofw', dev_path))

def read_ofw(dev_path):
    """Read a node in the OFW device tree.  Return None if node not present."""
    real_path = os.path.join('/ofw',dev_path)
    if not os.path.exists(real_path): return None
    with open(real_path) as f:
        return f.read()

def trim_ofw(ofw_contents):
    """Trim off the \n\0 on OFW mfg data device tree nodes.  According to
    OLPC trac #2085, this is a bug that will eventually be fixed -- but
    it doesn't hurt to try to strip them off in any case."""
    return ofw_contents.rstrip("\n\0") if ofw_contents is not None else None

###############################################################
## display prettiness
def clear_scr():
    print '\x1B[H\x1B[J' # clear screen, the hard way.

###############################################################
## activation support
def activation_fail (serial_num):
    """Display activation failure message, pause, then shutdown."""
    try:
        #  This message is never seen unless the GUI failed.
        print "Could not activate this XO."
        print "Serial number:", serial_num # don't show UUID
        # activation failed.  shutdown in 2 minutes.
        check_call(['/bin/sync'])
        time.sleep(60)
        poweroff()
    except:
        sys.exit(1)

def validate_activation (serial_number, uuid, lease_contents):
    """Check the validity of the given lease_contents."""
    from bitfrost.leases.core import find_lease
    try:
        find_lease(serial_number, uuid, lease_contents)
        return True
    except:
        return False

def check_stolen():
    # XXX: we should provide some way to delete the 'stolen' identifier
    # XXX: to recover the machine.
    if os.path.exists('/sysroot/.private/stolen'):
        # this machine is stolen!  delete activation lease
        try:
            os.unlink('/sysroot/security/lease.sig')
            fd = os.open('/sysroot/security', os.O_RDONLY)
            os.fsync(fd)
            os.close(fd)
        except: pass
        # shutdown now!
        poweroff()
    if not os.path.isdir('/sysroot/.private'):
        with writable_root():
            os.mkdir('/sysroot/.private')
    if not os.path.isdir('/sysroot/security/.private'):
        with writable_root():
            os.mkdir('/sysroot/security/.private')
    # mount OATC's private scratch space.
    check_call(['/bin/mount','--bind','/sysroot/.private','/sysroot/security/.private'])

########################################
# upgrade support

# frob the /current symlink to start from the backup filesystem if requested
# return the 'short name' for the image we should boot, or None if the
# filesystem is not upgradable.
def frob_symlink(boot_backup):
    if not os.path.islink('/sysroot/versions/boot/current'): return None
    # the 'current' symlink is of the form /versions/pristine/<hash>
    # we want to return <hash>
    p, current = os.readlink('/sysroot/versions/boot/current').rsplit('/', 1)
    assert p in ['/versions/pristine','/versions/run','../../run']
    if boot_backup:
        p, alt = os.readlink('/sysroot/versions/boot/alt').rsplit('/', 1)
        assert p in ['/versions/pristine','/versions/run','../../run']
        # atomically swap current and alt.
        from tempfile import mkdtemp
        with writable_root():
            config = os.readlink('/sysroot/versions/boot')
            d = mkdtemp(prefix='cfg.', dir='/sysroot/versions/configs')
            os.symlink(os.path.join('/versions/pristine', alt),
                       os.path.join(d, 'current'))
            os.symlink(os.path.join('/versions/pristine', current),
                       os.path.join(d, 'alt'))
            os.symlink(d.replace('/sysroot/versions/',''),
                       '/sysroot/versions/boot.tmp')
            check_call(['/bin/sync']) # superstition
            os.rename('/sysroot/versions/boot.tmp', '/sysroot/versions/boot')
            check_call(['/bin/sync']) # superstition
            # remove old config.
            check_call(['/bin/rm','-rf','/sysroot/%s' % config])
        current, alt = alt, current
    # check that /versions/run/$current exists; create if needed.
    if not os.path.isdir('/sysroot/versions/run/'+current):
        # steal setattr from root.
        root='/sysroot/versions/pristine/'+current
        if not os.path.exists('/usr/sbin/setattr') and \
               os.path.exists(root + '/usr/sbin/setattr'):
            check_call(['/bin/cp','%s/usr/sbin/setattr' % root,'/usr/sbin'])
        # create /versions/run/$current
        with writable_root():
            import upfs # upgrade fs helper code
            tree = upfs.Tree(current, mode='thawed') # frozen?
            tree.install('/sysroot')
    # create 'running' symlink
    def makelink(src, dst):
        try: # trac #5317: only recreate symlink if necessary.
            if os.readlink(dst) == src:
                return # it's already perfect, why change?
        except OSError:
            pass # dst not yet present.
        with writable_root():
            try:
                os.unlink(dst)
            except OSError:
                pass # dst not yet present.
            try:
                os.symlink(src, dst)
            except OSError:
                print "WARNING: Failed to create /versions/running." # nonfatal
    makelink('pristine/'+current, '/sysroot/versions/running')
    # done.
    return current

####################################
# Boot animation
boot_animation_started = False
def start_boot_animation(root='/sysroot'):
    global boot_animation_started
    if boot_animation_started: return
    if not os.path.exists(root+'/etc/init.d/0-boot-anim-start'): return
    boot_animation_started = True
    os.system('mount -t proc proc '+root+'/proc')
    os.system('mount -t sysfs sysfs  '+root+'/sys')
    # XXX might want to bind-mount /home here to allow early
    # bootanim customization
    with writable_root():
        for d in [('fb0','c 29 0'), ('console','c 5 1'),
                  ('tty1','c 4 1'), ('tty2','c 4 2')]:
            if not os.path.exists(root+'/dev/'+d[0]):
                os.system('mknod '+root+'/dev/'+d[0]+' '+d[1])
        os.system('/usr/sbin/chroot '+root+' /etc/init.d/0-boot-anim-start start')
    os.system('umount '+root+'/proc')
    os.system('umount '+root+'/sys')

###################################
# Special FRS upgrade code
def try_frs_upgrade(xo, bootpath):
    """Attempt to perform an automatic upgrade to the FRS build."""
    with root_mounted(xo, bootpath):
        if not os.path.exists('/sysroot/.first-boot'):
            return # don't do the upgrade
        treename = frob_symlink(False)
        if not treename: return # not upgradable?!
        from activate import setup_network, select_network_channel
        setup_network()
        time.sleep(1)
        select_network_channel(1)
        with writable_root():
            os.unlink('/sysroot/.first-boot')
            boot_run_xo(treename, frs_upgrade=True)


####################################
# Final boot
def boot_xo():
    """Free initramfs and transfer control to /sbin/init."""
    check_call(['/bin/umount','/proc'])
    check_call(['/bin/umount','/sys'])
    from runinit import runinit
    runinit('/sysroot')

def boot_run_xo(treename, frs_upgrade=False):
    # okay, do the boot!
    # hack: remount root rw before starting init.
    with writable_root():
        # create some bind mounts
        # (do this after making root writable, because newer kernels will
        #  otherwise end up with ro bind mounts against the writable root)
        for frag in [ 'home', 'security', 'versions' ]:
            # ignore failures here: a debian installation (say) may not have
            # these dirs.
            call(['/bin/mount','--bind','/sysroot/%s' % frag,
                  '/sysroot/versions/run/%s/%s' % (treename, frag)])
        check_call(['/bin/umount','/proc'])
        check_call(['/bin/umount','/sys'])
        from runinit import runinit
        runinit('/sysroot', '/versions/run/%s' % treename)
