""" run-init.pyx

 Usage: exec run-init [-c /dev/console] /real-root /sbin/init "$@"

 This program should be called as the last thing in a shell script
 acting as /init in an initramfs; it does the following:

 - Delete all files in the initramfs;
 - Remounts /real-root onto the root filesystem;
 - Chroots;
 - Opens /dev/console;
 - Spawns the specified init program (with arguments.)
"""
import os, os.path, stat, sys

# tiny bit of C for the mount() syscall.
cdef extern from "sys/mount.h":
    int mount(char *source, char *target, char *fstype, unsigned long flags, void *data)
    int umount(char *target)
cdef extern from "linux/fs.h":
    enum:
        MS_MOVE
# statfs, too.
cdef extern from "sys/vfs.h":
    struct statfs:
        long f_type
    int c_statfs "statfs"(char *path, statfs *buf)
# better exit than sys.exit
cdef extern from "unistd.h":
    void _exit(int status)
# these don't appear in glibc headers
TMPFS_MAGIC =	0x01021994
RAMFS_MAGIC =	0x858458f6

def nuke_dirent(directory, name, me):
    path = os.path.join(directory, name)
    st = os.lstat(path)
    if st.st_dev == me: # don't recurse down mount points!
        nuke(path)

def nuke_dir(what):
    """Wipe the contents of a directory, but not the directory itself."""
    st = os.lstat(what)
    if not stat.S_ISDIR(st.st_mode):
        raise OSError, "Not a directory"
    for name in os.listdir(what):
        nuke_dirent(what, name, st.st_dev)

def nuke(what):
    try:
        os.unlink(what)
    except OSError: # it's a directory?
        nuke_dir(what)
        os.rmdir(what)

def runinit(realroot, subroot='.', frs_upgrade=False):
    # first, change to the new root directory
    os.chdir(realroot)
    # this is potentially destructive: take precautions
    rst = os.stat('/')
    cst = os.stat('.')
    if rst.st_dev == cst.st_dev:
        raise OSError, "current directory on the same filesystem as the root"
    # the initramfs should have /init
    os.stat('/init')
    if not os.path.isfile('/init'):
        raise OSError, "can't find /init on initramfs"
    # make sure we're on a ramfs
    cdef statfs sfs
    if c_statfs('/', &sfs) != 0:
        raise OSError, "statfs failed."
    if sfs.f_type != RAMFS_MAGIC and sfs.f_type != TMPFS_MAGIC:
        raise OSError, "rootfs not a ramfs or tmpfs"

    # OK, I think we should be safe

    # delete rootfs contents
    nuke_dir('/')

    # Overmount the root -- all files are gone, so this can't be
    # a subprocess invocation
    if mount('.', '/', NULL, MS_MOVE, NULL) != 0:
        raise OSError, "mount --move failed."

    # chroot, chdir
    os.chroot('.')
    os.chdir('/')
    # do it again
    os.chdir(subroot)
    os.chroot('.')
    os.chdir('/')
    # change namespace.
    umount('/security/.private')
    if os.fork() != 0:
        _exit(0) # sys.exit raises an exception, which is the wrong thing
    # open /dev/console
    confd = os.open('/dev/console', os.O_RDWR)
    os.dup2(confd, 0)
    os.dup2(confd, 1)
    os.dup2(confd, 2)
    os.close(confd)
    # try to invoke python process
    if os.path.exists('/sbin/olpc_init.py'):
        try:
            # add /sbin to sys.path; olpc_init.run() should tidy up the
            # path once it starts.
            sys.path.insert(0, '/sbin')
            from olpc_init import run
            return run() # no parameters yet; this should be a tail call.
        except:
            print "FAILED TO START OLPC_INIT"
            print "Path:", sys.path
            print "Current dir:", os.getcwd()
            print "Exception:", sys.exc_info()
            print "Falling back to /sbin/init"
    # spawn init
    if frs_upgrade: # special support for MP-to-FRS upgrade
        print "Launching: /usr/sbin/olpc-update frs"
        os.execl('/usr/sbin/olpc-update','/usr/sbin/olpc-update','frs')
    else:
        os.execl('/sbin/init','/sbin/init','--init')
    raise OSError, "Should not get here!"
